#!/usr/bin/perl -w
# ****************************************************************************
# This is supposed to be a replacement for the mauireswww-tool that is
# distributed with maui in the contrib dir. Though it is generally a nice
# tool, using C for this purpose seems strange, and makes extensions hard
# and unmaintainable. It also isn't exectly easily cusomizable. After having
# handpatched mauireswww for two different clusters with several specialities
# and additional features, it was time for a rewrite. Here it is.
# ****************************************************************************

# Configuration. Note: These are emergency fallback defaults only.
# Put your configuration into a file called 'mauireswww-settings.pl'
# that you put somewhere in the include searchpath, preferably into the
# same dir as the script.

# Where showres (and showq) is
#$mauirdir='/opt/maui/current/bin/';
$mauidir='/usr/bin/';

# List of nodes. Can be a textfile with just one node per line. Can also
# be a pbs node file, in which case the 'np=' values are also used for
# showing <1 node jobs.
$nodelist='/var/spool/pbs/server_priv/nodes';

# How many days to show
$daystoshow = 15;

# Pixels per hour
$pixelsperhour = 10;

# Height of one node. Note: Should be divideable by any CPU count you have
# in your cluster machines if you want to use the splitnodes feature.
$nodeheight = 12;

# Prefix for the images for the graphs. Relative or absolute doesn't matter,
# just gets written to the HTML like that.
# '' is okay if pictures are in the same dir as the script.
$imgdir = '';

# Split nodes into CPUs so you see jobs with tasks < np properly?
# Warning: this requires a nodelist with 'np=X' to work! Also makes the
# output a lot bigger.
$splitnodes = 1;

# Cluster Name
$clustername = 'Some Cluster';

# Write HTML Header and Footer in standalone mode?
$headandfootinstandalone = 1;

# Show Reservation names?
$showresnames = 0;

# Reservation Coloring Rules. This is a list of regexps and their resulting
# color. The list is traversed until a match is found. If no match is found,
# the default is 'red'. However, it is recommended to end the list with a
# default entry '.*'.
@rescoloringrules = ( [ '^devel',         'yellow' ],
                      [ '^.{0,1}express', 'yellow' ],
                      [ '.*',             'red' ]
                    );

# Link to Ganglia Jobmonarch? Set to '' if you don't want such a link, or
# to the correct Ganglia URL if you want. The job ID will be inserted into
# that URL where you write '%ID%'.
$jobmonarchurl = 'http://someserver/ganglia/addons/job_monarch?c=yourcluster&id=%ID%';

# Show job status below the graphical representation
$showjobstatus = 1;

# Page Header and Footer - there is no default here, because some variables
# could not be set yet. The default is hidden inside the function
# writedefaulthtmlheader / writedefaulthtmlfooter.
$pageheader = undef;
$pagefooter = undef;

# ------------- End of fallback configuration ----------

# Do NOT change this unless you know what you're doing.
$mauiresstyles = <<EOSTYLES;
      p,th,td { font-family: Arial,Helvetica; }
      .nodename { font-size: 12px; text-align: right; }
      .nodealloc td { font-size: 12px; }
      .nodealloc th { font-style: bold; text-align: left; background-color: #ffffa0; }
      .reservations { font-size: 1px; white-space: pre; background-color: #ffffa0; }
      .legend { font-size: 14px; }
      .note { color: #800000 }
      .resimgs { border-color: #000000; border-style: solid; border-left-width: 0px; border-right-width: 1px; border-top-width: 0px; border-bottom-width: 0px; }
EOSTYLES

sub writedefaulthtmlheader {
  if (defined($pageheader)) {
    print($pageheader);
  } else {
    print <<EOHEAD;
  <html>
  <head>
    <title>$clustername Status</title>
    <style>
      /* mauireswww */
      $mauiresstyles
    </style>
    <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
  </head>
  <body>
    <h1>$clustername Status</h1>
EOHEAD
  }
  print(localtime(time())."");
  if ($showjobstatus) {
    print("<h2>Nodes and Reservations</h2>");
  }
}

sub writedefaulthtmlfooter {
  if (defined($pagefooter)) {
    print($pagefooter);
  } else {
    print <<EOFOOT;
  </body>
  </html>
EOFOOT
  }
}

sub loadshowres {
  my $sr; my $ll; my $hn; my %oneres;
  %reservations = ();
  unless (open($sr, $mauidir.'showres -n -r|')) {
    print("Failed to execute showres as '${mauidir}showres -n -r'.\n");
    return 0;
  }
  while ($ll = <$sr>) {
    $ll =~ s/[\r\n]//g;
    #   1           2                 3       4         5      6    7        8       9
    # w0830        Job              45324    Running    4   -66022 -->    20378  (00086400)
    if ($ll =~ m/^\s*([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s*$/) {
      $hn = $1;
      $oneres{'hostname'} = $1;
      $oneres{'type'} = lc($2);
      if (($oneres{'type'} ne 'job') && ($oneres{'type'} ne 'user')) {
        # Unknown type, ignore.
        next;
      }
      $oneres{'id'} = $3;  # Either a number or a name (SYSTEM.0)
      $oneres{'state'} = $4;
      $oneres{'tasks'} = int($5);
      $oneres{'start'} = $6;
      $oneres{'end'} = $8;
      $oneres{'duration'} = $9;
      $oneres{'duration'} =~ s/[()]//g;
      push(@{$reservations{$hn}}, {%oneres});
    }
    #print("$ll\n");
  }
  close($sr);
  return 1;  # OK
}

sub loadnodelist {
  my $nl; my $ll;
  my $i = 0;
  @nodes = ();
  unless (open($nl, '<'.$nodelist)) {
    print("Failed to load nodelist from '$nodelist'.\n");
    return 0;
  }
  while ($ll = <$nl>) {
    $ll =~ s/[\r\n]//g;
    if      ($ll =~ m/^([^ #\t\r\n]+).*np=(\d+)/) {
      $nodes[$i]{'name'} = $1;
      $nodes[$i]{'np'} = int($2);
      $i++;
    } elsif ($ll =~ m/^([^ #\t\r\n]+)/) {
      $nodes[$i]{'name'} = $1;
      $nodes[$i]{'np'} = -1;
      $i++;
    } else {
      # Ignore the line.
    }
  }
  close($nl);
  return 1;  # OK
}

# Parameter: Number of seconds
# Returns: Width in Pixels
sub scaletime ($) {
  return ($_[0] * $pixelsperhour) / 3600;
}

sub calculatereservations ($@) {
  my @resvs = @{$_[1]}; my $i; my $ncpus = $_[0]; my $j; my $run; my $c;
  @timetable = ();
  for ($run = 0; $run < 2; $run++) {
    # Two runs: first for user/system reservations, then jobs.
    for ($i = 0; $i < @resvs; $i++) {
      my %oneres = %{$resvs[$i]};
      if (($run == 0) && ($oneres{'type'} ne 'user')) { next; }
      if (($run == 1) && ($oneres{'type'} ne 'job')) { next; }
      if ($oneres{'end'} eq 'INFINITE') { $oneres{'end'} = 99999999999; }
      if ($oneres{'start'} eq '-INFINITY') { $oneres{'start'} = -1; }
      if ($oneres{'end'} < 0) { # Entry in the past? sod it.
        next;
      }
      if ($oneres{'start'} > $oneres{'end'}) { # End before start? WTF?
        next;
      }
      if ($oneres{'start'} < 0) { # It started in the past, but our graph starts now.
        $oneres{'start'} = 0;
      }
      if ($oneres{'end'} > ($daystoshow * 24 * 3600)) { # It goes further than our graph -> cut it.
        $oneres{'end'} = $daystoshow * 24 * 3600;
      }
      if (($oneres{'type'} eq 'user') && ($ncpus > 0)) {
        #print(STDERR "Correction tasks from $oneres{'tasks'} to $ncpus for $oneres{'id'}\n");
        $oneres{'tasks'} = $ncpus;
      }
      # Ok, so we now know how far the reservation goes. Lets find free line(s) for it.
      for ($j = scaletime($oneres{'start'}); $j < scaletime($oneres{'end'}); $j++) {
        my $curcpu = 0;
        for ($c = 0; $c < (($splitnodes) ? $oneres{'tasks'} : 1); $c++) { # One line for each task if we split nodes
          if ($splitnodes) {
            while (($curcpu < (($ncpus > 0) ? ($ncpus-1) : 31))  # Just for sanity, an emergency exit
                && (defined($timetable[$curcpu][$j])) # Not empty, but if it's a 
                && ($timetable[$curcpu][$j] !~ m/^user/) # userres it can be overwritten - jobs can't
                  ) {
              $curcpu++;
            }
          }
          $timetable[$curcpu][$j] = $oneres{'type'}.' '.$oneres{'state'}.' '.$oneres{'id'};
          if ($curcpu < (($ncpus > 0) ? ($ncpus-1) : 31)) {
            $curcpu++; # do not ever reuse our own spot.
          }
        }
      }
    }
  }
}

# Par. 0: Reservation name
# Returns: Filename of the image for the reservation
sub getimgfn($) {
  if ($_[0] eq 'UNDEF') {
    return 'transp.gif';
  }
  my @fields = split(/ /, $_[0]);
  if ($fields[0] eq 'user') {
    my $rc;
    foreach $rc (@rescoloringrules) {
      if ($fields[2] =~ m/$rc->[0]/i) {
        return $rc->[1].'.gif';
      }
    }
    return 'red.gif';
  }
  if (lc($fields[1]) eq 'running') {
    return 'darkgreen.gif';
  } else {
    return 'green.gif';
  }
}

# Par. 0: Reservation name
# Par. 1: what getimgfn returned (so we don't need the logic twice)
# Returns: Alt text for image for the reservation
sub getalttext($$) {
  if ($_[1] eq 'transp.gif') {
    return '';
  }
  my @fields = split(/ /, $_[0]);
  if (($_[1] eq 'green.gif') || ($_[1] eq 'darkgreen.gif') || ($showresnames)) {
    return $fields[2];
  } else {
    return '';
  }
}

# Par. 0: Reservation name
# Returns: nothing
# Prints the start (opening A tag) of a link to jobmonarch.
sub writejobmonarchlinkstart($) {
  if ((!defined($jobmonarchurl)) || ($jobmonarchurl eq '')) {
    return;
  }
  my @fields = split(/ /, $_[0]);
  if ((defined($fields[2])) && ($fields[2] =~ m/^\d+$/)) {
    my $changedurl = $jobmonarchurl;
    $changedurl =~ s/%ID%/$fields[2]/ge;
    print("<a href=\"$changedurl\">");
    $job_mon_close_tag = '</a>';
  }
}

# closes a job monarch link tag - note: requires status kept in global variable.
sub writejobmonarchlinkend() {
  if ((defined($job_mon_close_tag)) && ($job_mon_close_tag ne '')) {
    print($job_mon_close_tag);
    $job_mon_close_tag = undef;
  }
}

sub showdaytimeline () {
  # potential fixme: Assume dayimageheight of 12 and dayimagewidth of 60
  # because that is what the original mauireswww images use.
  unless (defined($dayimgh)) { $dayimgh = 12; }
  unless (defined($dayimgw)) { $dayimgw = 60; }
  my $k; my $width = 1;
  @dayfilenames = ( 'sunday.gif', 'monday.gif', 'tuesday.gif',
                    'wednesday.gif', 'thursday.gif', 'friday.gif',
                    'saturday.gif', 'sunday.gif'
                  );
  print("<tr><td class=\"nodename\">&nbsp;</td><td class=\"legend\">");
  my $lwkday = (localtime(time()))[6];
  my $dfn = 'transp.gif';
  for ($k = 0; $k < ($daystoshow * 24 * $pixelsperhour); $k++) {
    my $wkday = (localtime(time() + ($k * 3600) / $pixelsperhour))[6];
    if ($wkday == $lwkday) {
      $width++;
    } else {
      if ($width >= $dayimgw) {
        print("<img border=\"0\" height=\"$dayimgh\" width=\"$dayimgw\" TITLE=\"$dfn\" NAME=\"$dfn\" ALT=\"$dfn\" src=\"${imgdir}${dfn}\">");
        $width -= $dayimgw;
      }
      if ($width > 0) {
        print("<img border=\"0\" height=\"$dayimgh\" width=\"$width\" TITLE=\"$dfn\" NAME=\"$dfn\" ALT=\"$dfn\" src=\"${imgdir}transp.gif\">");
      }
      $width = 1;
      $lwkday = $wkday;
      $dfn = $dayfilenames[$wkday];
    }
  }
  if ($width >= $dayimgw) {
    print("<img border=\"0\" height=\"$dayimgh\" width=\"$dayimgw\" TITLE=\"$dfn\" NAME=\"$dfn\" ALT=\"$dfn\" src=\"${imgdir}${dfn}\">");
    $width -= $dayimgw;
  }
  if ($width > 0) {
    print("<img border=\"0\" height=\"$dayimgh\" width=\"$width\" TITLE=\"$dfn\" NAME=\"$dfn\" ALT=\"$dfn\" src=\"${imgdir}transp.gif\">");
  }
  print("</td></tr>\n");
}

sub printshowq {
  unless (open($sq, $mauidir.'showq|')) {
    print("Failed to execute showres as '${mauidir}showq'.\n");
    return 0;
  }
  my $ll; my $preisopen = 1;
  while ($ll = <$sq>) {
    if ($ll =~ m/^ACTIVE JOBS/) {
      if ($preisopen) { print("</pre>\n"); }
      print("<h4>Active Jobs</h4><pre>");
      $preisopen = 1;
      next;
    }
    if ($ll =~ m/^IDLE JOBS/) {
      if ($preisopen) { print("</pre>\n"); }
      print("<h4>Idle Jobs</h4><pre>");
      $preisopen = 1;
      next;
    }
    if ($ll =~ m/^BLOCKED JOBS/) {
      if ($preisopen) { print("</pre>\n"); }
      print("<h4>Blocked Jobs</h4><pre>");
      $preisopen = 1;
      next;
    }
    if ($ll =~ m/^JOBNAME/) {
      print("<b>$ll</b>");
      next;
    }
    print($ll);
  }
  if ($preisopen) { print("</pre>\n"); }
  close($sq);
}

# --- main()

# Try to load settings first
eval {
  require('mauireswww-settings.pl');
};
if (defined($ENV{'REQUEST_METHOD'})) { # CGI mode
  $|=1; # Autoflush, jaja
  print("Content-type: text/html\r\n\r\n");
} else { # Standalone

}
if ((defined($ENV{'REQUEST_METHOD'})) || ($headandfootinstandalone)) {
  writedefaulthtmlheader();
}
unless (loadnodelist()) {
  print("Nothing to show without a nodelist. Abort.\n");
  exit(0);
}
unless (loadshowres()) {
  print("Reservations could not be loaded.\n");
  exit(0);
}
print("<table class=\"nodealloc\" cellpadding=\"0\" cellspacing=\"0\" border=\"0\">");
showdaytimeline();
for ($j = 0; $j < @nodes; $j++) {
  $hn = $nodes[$j]{'name'};
  if (defined(@{$reservations{$hn}})) {
    @resvs = @{$reservations{$hn}};
  } else {
    @resvs = ();
  }
  calculatereservations($nodes[$j]{'np'}, \@resvs);
  $ncpus = int(@timetable);
  if ($ncpus < 1) {
    $ncpus = 1;
  }
  if (($splitnodes) && ($ncpus < $nodes[$j]{'np'})) {
    $ncpus = $nodes[$j]{'np'};
  }
  print("<tr><td class=\"nodename\">$hn</td><td class=\"reservations\">");
  my $imgfn;
  my $imgh = $nodeheight / $ncpus;
  for ($c = 0; $c < $ncpus; $c++) {
    $lastres = $timetable[$c][0]; $width = 1;
    for ($k = 0; $k < ($daystoshow * 24 * $pixelsperhour); $k++) {
      if (!defined($lastres)) { $lastres = 'UNDEF'; }
      if (!defined($timetable[$c][$k])) { $timetable[$c][$k] = 'UNDEF'; }
      if ($timetable[$c][$k] eq $lastres) {
        $width++;
      } else {
        $imgfn = getimgfn($lastres);
        $imgalt = getalttext($lastres, $imgfn);
        writejobmonarchlinkstart($lastres);
        #$width -= 1; # Needed in case of border tricks (use class=resimgs instead of border=0)
        print("<img border=\"0\" height=\"$imgh\" width=\"$width\" NAME=\"$imgalt\" TITLE=\"$imgalt\" ALT=\"$imgalt\" src=\"${imgdir}${imgfn}\">");
        writejobmonarchlinkend();
        $width = 1;
        $lastres = $timetable[$c][$k];
      }
    }
    $imgfn = getimgfn($lastres);
    $imgalt = getalttext($lastres, $imgfn);
    writejobmonarchlinkstart($lastres);
    print("<img border=\"0\" height=\"$imgh\" width=\"$width\" NAME=\"$imgalt\" TITLE=\"$imgalt\" ALT=\"$imgalt\" src=\"${imgdir}${imgfn}\">");
    writejobmonarchlinkend();
    print("<br>");
  }
  print("</td></tr>\n");
}
showdaytimeline();
print("</table>\n");
if ($showjobstatus) {
  print("<hr noshade>\n");
  print("<h2>Job Status</h2>");
  printshowq();
}
if ((defined($ENV{'REQUEST_METHOD'})) || ($headandfootinstandalone)) {
  writedefaulthtmlfooter();
}
